'''/network/tn.py
Implementation of the Topic-aware Network model
'''

import os
import random
import pathlib
import collections
import networkx as nx
import numpy as np
from scipy import sparse
from scipy.spatial import distance
from tqdm import tqdm, tqdm_notebook
from sklearn.decomposition import NMF
import womg_core
from womg_core.network.tlt_network_model import TLTNetworkModel
from womg_core.utils.utility_functions import read_edgelist


DEFAULT_GRAPH = pathlib.Path(os.path.abspath(womg_core.__file__).replace('/womg/__init__.py',
                             ''))/ "womgdata" / "graph" / "lesmiserables" / "lesmiserables_edgelist.txt"

class TN(TLTNetworkModel):
    '''
    Implementation of interest-influence model

    Attributes
    ----------
    - users_interests : dict
        dictionary containing the  interests vectors
        (of _numb_topics dimension) in the format:
        key <- [node id]
        value <- _numb_topics dimension array in numpy format
    - users_influence : dict
        dictionary containing the  influence vectors
        (of _numb_topics dimension) in the format:
        key <- [node id]
        value <- _numb_topics dimension array in numpy format
    - nx_obj : NetworkX object
        networkx instance of the input network
    - _numb_topics : int
        dimension of the interests and influence vectors
    - _fast : bool
        flag for defining the chosen method for interests generation.
        if True the fastest (random) method is chosen

    Methods
    -------
    - set_interests()
    - set_influence()
    - node2interests(): generates realistic latent interests of the nodes starting
      from the fixed given network
    - graph_weights_vecs_generation(): updates network weights using interests
      and influence vecs

    Notes
    -----
    This class implementation inherits from the abstract class networkModel()

    References
    ----------

    '''

    def __init__(self, numb_topics, #pylint: disable=too-many-arguments,
                 homophily,
                 weighted, directed,
                 graph,
                 interests,
                 gn_strength,
                 infl_strength,
                 progress_bar,
                 seed,
                 ):
        super().__init__()
        self._graph = graph
        self._weighted = weighted
        self.directed = directed
        self.users_interests = {}
        self.users_influence = {}
        self._numb_topics = numb_topics
        self._homophily = homophily
        self._interests = interests
        assert gn_strength is None or np.isfinite(gn_strength)
        self._godnode_strength = gn_strength
        self._infl_strength = infl_strength
        self._rand = 16 - 15.875 * self._homophily
        if progress_bar:
            self._progress_bar = tqdm_notebook
        else:
            self._progress_bar = tqdm
        self._verbose = False
        self._seed = seed
        self.nx_obj = None
        self.mapping = None




    def network_setup(self, int_mode):
        '''
        - Sets the interests vectors using set_interests() method
        - Sets the influence vectors using set_influence() mehtod
        - Sets the new graph weights using update_weights() method

        Notes
        -----
        See each method docstring for details
        '''
        if isinstance(self._graph, nx.classes.graph.Graph):
            self.nx_obj = self._graph.copy()
            self.directed = self.nx_obj.is_directed()
            if not self.directed:
                self.nx_obj = self.nx_obj.to_directed()
            self._weighted = nx.is_weighted(self.nx_obj)
        elif self._graph is None:
            print('No graph path provided \n',
                  'DEMO Mode: generating cascades in les miserables network')
            self._graph = DEFAULT_GRAPH
            self.nx_obj, self.mapping = read_edgelist(path=self._graph,
                                                      weighted=False,
                                                      directed=False)
        else:
            self._graph = pathlib.Path(self._graph)
            self.nx_obj, self.mapping = read_edgelist(path=self._graph,
                                                      weighted=self._weighted,
                                                      directed=self.directed)
        if self._godnode_strength is not None:
            self.set_godnode_links()
        self.set_interests(int_mode)
        ##### check inf ##########################################
        for node, interests in self.users_interests.items():
            if not all(np.isfinite(interests)):
                print('ATTENTION  node: ', node, '  interests: ', interests)
        ###########################################################
        self.set_influence()
        self.graph_weights_vecs_generation()
        if self._verbose:
            print('Macroscopic homophily level: ',
                  self.homophily(),
                  ' with H=', self._homophily)



    def set_godnode_links(self, weight=1, nodes=None):
        '''
        Sets the godNode's links weights in the graph

        Parameters
        ----------
        weight : int
            scalar weight for the link/links
            (Default 1)

        nodes : iterable/int
            nodes indexes or node index on which one wants to change link weights
            if None is given all link weights (from godNode to each node) is changed
            (Defalut None)

        Example:
        tn_instance.set_godnode_links() :
        all the links weights (from godNode to each node) are set to 1
        '''
        if nodes is None:
            #print('Setting god node')
            god_node_links = []
            for node in self._progress_bar(self.nx_obj.nodes()):
                rand_num = np.abs(np.random.randn())
                god_node_links.append((-1, node, rand_num))
            self.nx_obj.add_weighted_edges_from(god_node_links)
        if isinstance(nodes, int):
            self.nx_obj.add_weighted_edges_from([(-1, nodes, weight)])
        if isinstance(nodes, collections.Iterable):
            for node in self._progress_bar(nodes):
                self.nx_obj.add_weighted_edges_from([(-1, node, weight)])


    def set_interests(self, int_mode):
        '''
        Creates interests vectors (numb_topics dimension) for each node and
        save them in the users_interests class attribute.

        Parameters
        ----------
        - method : string
          name of the method for creating interests vectors
        '''
        print('Creating interests..')
        if int_mode == 'rand':
            #print('Random generation of interests:')
            self.random_interests()

        if int_mode == 'nmf':
            self.nmf_interests()

        if int_mode == 'load' or self._interests is not None:
            if isinstance(self._interests, dict):
                print('Loading interests')
            else:
                print('Loading interests from: ', self._interests)
            self.load_interests()


    def set_influence(self):
        '''
        Sets influence vectors for all nodes

        Parameters
        ----------
        - method : string
          name of the method for creating interests vectors
        '''

        if self._infl_strength is None:
            for node in self.nx_obj.nodes():
                self.users_influence[node] = [0. for _ in range(self._numb_topics)]
        else:
            fitness = 1 + np.random.pareto(a=self._infl_strength,
                                           size=self.nx_obj.number_of_nodes())
            if np.any(np.isinf(fitness)):
                print('Wrong parameter infl strenght')
            fitness = np.minimum(fitness, 10E20)
            for node in self.nx_obj.nodes():
                self.users_influence[node] = fitness[node] * self.users_interests[node]

        if self._godnode_strength is not None:
            self.users_influence[-1] = np.ones(self._numb_topics)

        for node, value in self.users_influence.items():
            if not all(np.isfinite(value)):
                print('ATTENTION node: ', node, ' influences: ', value)


    def random_interests(self, norm=True):
        '''
        Create interests vector for each node using a random uniform probability density
        function and directly saves interests vectors in attribute users_interests

        Parameters
        ----------
        - norm : bool
        if True interests vectors are normalized

        '''
        if norm:
            for node in self.nx_obj.nodes():
                self.users_interests[node] = np.random.rand(self._numb_topics)

    @staticmethod
    def overlap_generator(A):
        """
        Generate the second order overlap from a sparse adjacency matrix A.
        """
        aat = A.dot(A.T)
        d = aat.diagonal()
        ndiag = sparse.diags(d, 0)
        n = np.sqrt(ndiag.dot(aat > 0).dot(ndiag))
        n.data[:] = 1./n.data[:]
        return aat.multiply(n) #- sparse.identity(aat.shape[0])



    def nmf_interests(self, eta=64.):
        '''
        Generates nodes interests using NMF
        '''
        #beta = self._homophily
        A = nx.adjacency_matrix(self.nx_obj)
        S_0 = self.overlap_generator(A)
        R = np.random.rand(self.nx_obj.number_of_nodes(), self.nx_obj.number_of_nodes())
        #S = beta*(S_0 + A + sparse.identity(A.shape[0])) + (1-beta)*R
        eta = 64.
        S = eta*S_0 + A + self._rand*R
        model = NMF(n_components=self._numb_topics, init='nndsvd', random_state=self._seed)
        W = model.fit_transform(S)
        if not np.all(np.isfinite(W)):
            print('ATTENTION W contains infinites')

        for node in self.nx_obj.nodes():
            self.users_interests[node] = W[node]

    def node2influence(self, scale_fact, alpha_max=10):
        '''
        Creates influence vectors (numb_topics dimension) from dirichlet
        distribution over topics for a single node

        Parameters
        ----------
        - alpha_max : int
          highest value in the dirichlet weight vec for a random entry
        '''
        dirich_weight_vec = [1 for topic in range(self._numb_topics)]
        dirich_weight_vec[random.randint(0, self._numb_topics-1)] = alpha_max
        influence_vec = np.random.dirichlet(dirich_weight_vec)
        return scale_fact*influence_vec


    def graph_weights_vecs_generation(self):
        '''
        Creates weights vectors (numb_topcis dimension) for each direct link
        and update the graph attribute; a link weight is defined as:

        w_(i,j) = intersts_j + influence_i

        Parameters
        ----------
        - god_node_weight : int
          scalar value for each entry of the weight vector involving the godNode
        '''
        for u, v in list(self.nx_obj.edges()):
            # god node
            if u == -1:
                out_influence_vec = self.users_influence[u]
                in_interest_vec = self.users_interests[v]
                self.set_link_weight((u, v),
                                     in_interest_vec + self._godnode_strength * out_influence_vec)
            else:
                out_influence_vec = self.users_influence[u]
                in_interest_vec = self.users_interests[v]
                self.set_link_weight((u, v), out_influence_vec + in_interest_vec)


    def set_link_weight(self, link, new_weight):
        '''
        Sets the link attribute to the given weights vector arg

        Parameters
        ----------
        - link : tuple
          link in the graph in which you want to change the attribute
        - new_weight : array
          numb_topic dimension array that is going to be the new attribute of the
          link
        '''
        u, v = link
        if not all(np.isfinite(new_weight)):
            print('ATTENTION link ', link, ' weight: ', new_weight)
        self.nx_obj[u][v]['weight'] = new_weight


    def load_interests(self, sep=','):
        '''
        Loads interests vector from path.
        Format: "node int1,int2,int3.."
        '''
        if isinstance(self._interests, dict):
            self.users_interests = self._interests
        else:
            with open(self._interests, 'r') as file:
                for line in file.readlines():
                    node, interests = line.split(' ')[0], line.split(' ')[1]
                    node_interests = [float(entry) for entry in interests[:-1].split(sep)]
                    self.users_interests[int(node)] = np.array(node_interests)


    ################# Analysis

    def sim_in(self):
        '''Returns average interests similarity between connected nodes
        '''
        sims = []
        for i in self.nx_obj.nodes:
            for j in list(self.nx_obj.neighbors(i)):
                sims.append(1 - distance.cosine(self.users_interests[i], self.users_interests[j]))
        return np.mean(sims)

    def select_notedge(self):
        '''Returns tuple of not connected nodes
        '''
        node1 = np.random.choice(self.nx_obj.nodes())
        node2 = np.random.choice(self.nx_obj.nodes())

        while (node1, node2) in self.nx_obj.edges or node1 == node2:
            node1 = np.random.choice(self.nx_obj.nodes())
            node2 = np.random.choice(self.nx_obj.nodes())
        return node1, node2

    def sim_out(self, samples):
        '''Returns average interests similarity between of not connected nodes
        samples times
        '''
        sims_out = []
        for _ in range(samples):
            i, j = self.select_notedge()
            sims_out.append(1 - distance.cosine(self.users_interests[i], self.users_interests[j]))
        return np.mean(sims_out)

    def homophily(self, numb_not_edges_tested=10000):
        '''Returns the cosine similarity homophily ratio measure
        '''
        return self.sim_in() / self.sim_out(numb_not_edges_tested)



'''/Topic/TopicModel.py
Abstract class defining Topic model
'''
import abc
from womg_core.topic.topic_model import TopicModel


class TLTTopicModel(TopicModel):
    '''
    Abstract class for topic models involved in tlt diffusion model class

    Attributes
    ----------
    viralities : dict
        key: document index, value: int virality parameter

    Methods
    -------
    fit : abstract
        inferes/generates topic distribution for each document

    set_docs_viralities : abstract
        sets the documents viralities
    '''

    def __init__(self):
        super().__init__()
        self.viralities = {}

    @abc.abstractmethod
    def fit(self):
        '''Methods for infering/generating topic distributions of the given documents
        '''
        pass

    @abc.abstractmethod
    def set_docs_viralities(self, virality_exp):
        '''Method for setting the documents viralities
        '''



'''Utility functions
'''
import pathlib
import pickle
import networkx as nx


def read_edgelist(path, weighted, directed):
    '''Reads the input network in networkx.
    '''
    if weighted:
        graph = nx.read_edgelist(path, nodetype=int,
                                 data=(('weight', float),),
                                 create_using=nx.DiGraph())
    else:
        graph = nx.read_edgelist(path, nodetype=int, create_using=nx.DiGraph())
        for edge in graph.edges():
            graph[edge[0]][edge[1]]['weight'] = 1

    if not directed:
        graph = graph.to_directed()

    # mapping labels
    mapping = {}
    identity_map = 0
    for new_label, old_label in enumerate(sorted(graph.nodes())):
        if new_label == old_label:
            identity_map += 1
        mapping[old_label] = new_label
    if identity_map == graph.number_of_nodes():
        return graph, None
    return nx.relabel_nodes(graph, mapping), mapping


def cleaning():
    '''Removing temporary class files

    Notes
    -----
    Temporary files start with '__'
    '''
    net_file = pathlib.Path.cwd() / "__network_model"
    if pathlib.os.path.exists(net_file):
        pathlib.os.remove(net_file)
    top_file = pathlib.Path.cwd() / "__topic_model"
    if pathlib.os.path.exists(top_file):
        pathlib.os.remove(top_file)

def count_files(path):
    '''Returns number of files in a directory; files that start wtih '.' are not
       considered

    Parameters
    ----------
    path : string
        directory path in which we want to know the number of files

    Notes
    -----
    It is not recursive: it does not count files in inner directories
    '''
    files = next(pathlib.os.walk(str(path)))[2]
    for file in files:
        if file.startswith('.'):
            files.remove(file)
    return len(files)

def read_graph(file_path):
    '''
    Reads graph from a path
    '''
    with open(file_path, 'rb') as file:
        rfile = pickle.load(file)
    return rfile

def read_docs(path, verbose=False):
    '''
    Returns the list of items, each row is the doc list

    Parameters
    ----------
    path : str
        position of the folder

    Returns
    -------
    docs in a list format : each entry of the list is a file in a list of words
    '''
    onlyfiles = [f for f in pathlib.os.listdir(path) if pathlib.os.path.isfile(pathlib.os.path.join(path, f))]
    onlyfiles = [f for f in onlyfiles if not f.startswith('.')]
    docs = []
    if verbose:
        print(onlyfiles)
    for file in onlyfiles:
        f_path = pathlib.Path(path) / str(file)
        with open(f_path, 'rb') as file:
            doc_list = [j for j in file]
            if verbose:
                print(doc_list)
            docs.append(doc_list)
    return docs


def find_numb_nodes(graph):
    '''
    Finds out number of nodes for a formatted graph
    '''
    maxx = 0
    for key in graph.keys():
        for i in range(2):
            if key[i] > maxx:
                maxx = key[i]
    return maxx



'''/network/tlt_network_model.py
Abstract class defining the Network models
'''

import abc
from womg_core.network.network_model import NetworkModel


class TLTNetworkModel(NetworkModel):
    '''
    Abstract class for network models involved in the tlt diffusion model class


    Methods
    -------
    graph_weights_vecs_generation : abstract
        method for generating numb_topics dim vectors for each link
    validate_network_config : concrete
        validate the configuration of the class model for the tlt class

    '''

    @abc.abstractmethod
    def graph_weights_vecs_generation(self):
        '''Generates numb_topics dim vectors for each link and put them as
        value of the graph dict attribute of the class
        '''
        pass

    @abc.abstractmethod
    def set_godnode_links(self, weight, nodes):
        '''Creates the god node links for initial configurations
        '''
        pass



'''/Topic/lda.py
Implementation of LDA topic-model
'''
import os
import pathlib
import numpy as np
from tqdm import tqdm, tqdm_notebook
import womg_core
from womg_core.topic.tlt_topic_model import TLTTopicModel
from womg_core.utils.distributions import random_powerlaw_vec


class LDA(TLTTopicModel):
    '''
    Class implementing Latent Dirichlet Allocation as topic model
    for topic distribution involved in tlt class model

    Attributes
    ----------
    numb_topics : int
        hidden
    items_keyw : dict
        dict of items in bow format

    Methods
    -------
    set_lda_mode : concrete
        sets the lda mode: reading or generating mode
    '''
    def __init__(self, numb_topics,
                 numb_docs,
                 items_descr,
                 progress_bar):
        super().__init__()
        self.numb_topics = numb_topics
        self.numb_docs = numb_docs
        self.numb_words = 20
        self._docs_path = None
        self._items_descr = items_descr
        self.items_keyw = {}
        self.dictionary = []
        self.main_data_path = pathlib.Path(os.path.abspath(womg_core.__file__).replace('/womg_core/__init__.py', ''))/'womgdata'
        self._training_path = self.main_data_path /'docs'/'training_corpus_ap'
        self._topics_descr_path = self.main_data_path / 'topic_model' / 'Topics_descript.txt'
        self.topics_descript = self.load_topics_descr(self._topics_descr_path)
        if progress_bar:
            self._progress_bar = tqdm_notebook
        else:
            self._progress_bar = tqdm


    def fit(self):
        '''
        Pipeline for fitting the lda model

        1. define the lda mode : generative mode / reading mode
        2. train lda
        3. get the items descriptions (topic distribution for each item)
        4. get the items keywords (bow list for each item)
        '''
        mode = self.set_lda_mode()
        if mode == 'load':
            if isinstance(self._items_descr, dict):
                self.items_descript = self._items_descr
                self.numb_docs = len(self._items_descr.keys())
            else:
                self.items_descript, self.numb_docs = self.load_items_descr(self._items_descr)
        if mode == 'gen':
            self.gen_items_descript()

    @staticmethod
    def load_topics_descr(path):
        '''
        Loads a topic description file (for each topic word distribution)
        '''
        with open(path, 'r') as file:
            topics_descript = file.readlines()
        return str(topics_descript[0])

    def set_lda_mode(self):
        '''
        Sets how lda has to work:
        reading a document folder or generating documents


        Parameters
        ----------
        path : string
            position of the document folder

        Returns
        -------
        reading : bool
            if True: it will read docs inside the given folder path or input folder
            if False: it will use lda for generating docs
        '''
        # setting mode
        if self.numb_docs is None and self._docs_path is None:
            mode = 'load'
            if self._items_descr is None:
                # pre-trained topic model with 15 topics and 50 docs
                self._items_descr = self.main_data_path / 'topic_model' / 'Items_descript.txt'
            else:
                pass
            if isinstance(self._items_descr, dict):
                print('Loading items descriptions (topic distrib for each doc)')
            else:
                print('Loading items descriptions (topic distrib for each doc) in: ',
                      self._items_descr)
        if self.numb_docs is not None and self._docs_path is None and isinstance(self._items_descr, dict):
            mode = 'load' # when womg extended has been executed in gen mode

        elif self.numb_docs is None and self._docs_path is not None and self._items_descr is None:
            print('Please install the womg extended version')

        elif self.numb_docs is not None and self._docs_path is not None and self._items_descr is None:
            print('Please install the womg extended version')

        elif self.numb_docs is not None and self._docs_path is None and self._items_descr is None:
            mode = 'gen'
            print('Setting LDA in generative mode: ',
                  self.numb_docs, ' documents, with ',
                  self.numb_topics, ' topics.')
            print('Training the LDA model ..')

        return mode


    def set_docs_viralities(self, virality_exp):
        '''
        Sets the documents viralities to the given scalar/vector

        Parameters
        ----------
        viralitiy : float
            Exponent of the pareto distribution for documents
            viralities.
        '''
        viralities = random_powerlaw_vec(gamma=virality_exp, dimensions=self.numb_docs)

        if np.size(viralities) == self.numb_docs:
            for item in range(self.numb_docs):
                self.viralities[item] = viralities[item]
        if np.size(viralities) == 1:
            for item in range(self.numb_docs):
                self.viralities[item] = viralities[0]

    def gen_items_descript(self):
        '''
        Generates the topic distribution for each item
        and stores it in the items_descript attribute
        '''
        print('generating items descript')
        alpha = [1.0 / self.numb_topics for i in range(self.numb_topics)]
        gammas = {}
        for item in range(self.numb_docs):
            gammas[item] = np.random.dirichlet(alpha)
        self.items_descript = gammas


    def load_items_descr(self, path):
        '''
        Returns the items_descript loaded from a file in path

        Parameters
        ----------
        path : int
            path of the items_descript file

        Returns
        -------
        tuple of items_descript loaded from path and numb_docs
        '''
        items_descr_dict = {}
        with open(path, 'r') as file:
            numb_docs = 0
            for line in file:
                line = line.replace(',', '').replace(']', '').replace('[', '')
                values = line.split()
                node = int(values[0])
                interests_vec = [float(i) for i in values[1:]]
                if self.numb_topics != len(interests_vec):
                    print("Please write the correct number of topics",
                          "as input or in case you give the items_descr_path",
                          " you can omit it")
                items_descr_dict[node] = interests_vec
                numb_docs += 1
        return items_descr_dict, numb_docs



'''
/network/network_model.py
Abstract class defining the Network models
'''
import abc


class NetworkModel(abc.ABC):
    '''
    Abstract class for network models

    Attributes
    ----------
    - graph : dict
    - info : dict

    Methods
    -------
    gformat : static
        formats the graph as dict from networkx format
    save_model_attr : concrete
        saves all the attributes
    save_model_class : concrete
        saves all the class in pickle file

    Notes
    -----
    Hidden attributes names has to start with "Hidden_"
    '''

    def __init__(self):
        self.graph = {}
        self.info = {}


    @staticmethod
    def gformat(nx_obj, directed=False):
        '''
        Formatting NetworkX graph to a Python dictionary format

        Parameters
        ----------
        - nx_obj : networkx object
          NetworkX instance of the graph

        Returns
        -------
        - [dict] Python3 dictionary of the graph where:
          key <- tuple which describes the link (node_1, node_2)
          value <- int weight of the link
        '''
        graph = {}
        print('Formatting graph..')
        if directed:
            for edge in nx_obj.edges():
                graph[(edge[0], edge[1])] = 1
        else:
            for edge in nx_obj.edges():
                graph[(edge[0], edge[1])] = 1
                graph[(edge[1], edge[0])] = 1

        return graph


    @abc.abstractmethod
    def network_setup(self, int_mode):
        '''Setting up the network environment for diffusion
        '''
        pass
